/************************************************************************
** MODULE INFORMATION **
************************
** FILE NAME:          netkern.c
** SYSTEM NAME:        Netmon
** TYPE NAME:          kernel access
** ORIGINAL AUTHOR(S): L. Th. Weerpas & W. Haverhals
** VERSION NUMBER:     1.0
** CREATION DATE:      1992/6/9
**
** DESCRIPTION:        kernel and timer access routines for SunOs 4.x
*************************************************************************
** CHANGES INFORMATION **
*************************
** REVISION:           $Revision$
** WORKFILE:           $Workfile$
** LOGINFO:            $Log$
*************************************************************************
** COPYRIGHT           (c) 1992
**                     L. Th. Weerpas & W. Haverhals
**                     All rights reserved.
**
** NOTICE              Redistribution  and use in source and binary forms
**                     are  permitted  provided  that the above copyright
**                     notice  and  this  paragraph are duplicated in all
**                     such forms and that any documentation, advertising
**                     materials,  and  other  materials  related to such
**                     distribution and use acknowledge that the software
**                     was  developed  by  above  mentioned  authors. The
**                     name(s)  of the authors may not be used to endorse
**                     or  promote  products  derived  from this software
**                     without  specific  prior  written permission.
**                     THIS  SOFTWARE  IS  PROVIDED ``AS IS'' AND WITHOUT
**                     ANY  EXPRESS  OR  IMPLIED  WARRANTIES,  INCLUDING,
**                     WITHOUT  LIMITATION,  THE  IMPLIED  WARRANTIES  OF
**                     MERCHANTIBILITY   AND  FITNESS  FOR  A  PARTICULAR
**                     PURPOSE.
************************************************************************/

#define _NETKERN_C

#include "netkern.h"
#include "netmon.h"

/*
 * GLOBAL DATA
 */

extern int           displaywait;

kvm_t              *kvmH;                 /* handle */
double              timebegin;
double              timenew;
double              timeold;
double              timedone;

double              kvmreadbytes  = 0;
long                kvmreaderrors = 0;

static int          kvmF;

static KVMOFST      ofst_time;
static KVMLIST      symlist[] =
{
    { "_time" },                        /* time in sec.usec since boot */
    { NULL }
};
#define NSYM (sizeof(symlist)/sizeof( KVMLIST ))

/*****************************************************************
** NAME:        KvmOpen                                     NETMON
** SYNOPSIS:    void KvmOpen
**              (
**                void
**              )
** RELATED TO:  Kvm..()
** DESCRIPTION: open the kernel for reading
** REMARKS:     - exits on error
**              - fills the global kernel access handle 'kvmH'
** RETURN TYPE: void
** EXIT STATUS: -
*****************************************************************/

void KvmOpen( void )
{
    kvmH = kvm_open( NULL, NULL, NULL, NULL, O_RDONLY, prog );
    if ( kvmH == NULL )
        eprintf( "%s: couldn't open VMUNIX for reading\n", prog );

    kvmF = open( "/dev/kmem", O_RDONLY );
    if ( kvmF < 0 )
        eprintf( "%s: couldn't open /dev/kmem for reading\n", prog );
}

/*****************************************************************
** NAME:        KvmNlist                                    NETMON
** SYNOPSIS:    void KvmNlist
**              (
**                KVMLIST *pnl, int length, char *name
**              )
** RELATED TO:  Kvm..()
** DESCRIPTION: retrieve the address offsets within the kernel
**              for the specified symbols in the KVMLIST array
**              pointed to by 'pnl' with length 'length'.
** REMARKS:     - exits on error
**              - checks if all requested symbols are properly
**                filled in.
** RETURN TYPE: void
** EXIT STATUS: -
*****************************************************************/

void KvmNlist( KVMLIST *pnl, int length, char *name )
{
    KVMLIST *plist;
    int      r;
    int      notall = 0;

    /*
     * collect all kernel addresses indicated in nlist
     */

    r = kvm_nlist( kvmH, pnl );
    if ( r != 0 )
        eprintf( "%s: bad kvm_nlist() in <%s>\n", prog, name );

    /*
     * check nlist on successful offset fill
     */

    for ( plist=pnl; plist->n_name!=NULL; plist++ )
    {
        if ( plist->n_type == 0 )
        {
            fprintf( stderr, "%s: unresolved symbol '%s' in <%s>\n",\
                     prog, plist->n_name, name );
            notall = 1;
        }
    }
    if ( notall )
        SysClose( 1 );
}

/*****************************************************************
** NAME:        KvmClose                                    NETMON
** SYNOPSIS:    void KvmClose
**              (
**                void
**              )
** RELATED TO:  Kvm..()
** DESCRIPTION: close the kernel after all reads
** REMARKS:     - uses the global kernel access handle 'kvmH'
** RETURN TYPE: void
** EXIT STATUS: -
*****************************************************************/

void KvmClose( void )
{
    kvm_close( kvmH );
    close( kvmF );
}

/*****************************************************************
** NAME:        KvmRead                                     NETMON
** SYNOPSIS:    void KvmRead
**              (
**                KVMOFST addr,     offset in the kernel space
**                void *buf,        user space recipient
**                unsigned nbytes   block length in bytes
**              )
** RELATED TO:  Kvm..()
** DESCRIPTION: transfer the contents of a kernel memory block
**              to user space. 'addr' must be a value retrieved
**              by means of KvmNlist().
** REMARKS:     - uses the global kernel access handle 'kvmH'
** RETURN TYPE: void
** EXIT STATUS: -
*****************************************************************/

void KvmRead( KVMOFST addr, void *buf, unsigned nbytes )
{
    if ( kvm_read( kvmH, addr, (char*) buf, nbytes ) != nbytes )
        eprintf( "bad kvm_read> %s\n", sys_errlist[errno] );

    kvmreadbytes += (double) nbytes;
}

/*****************************************************************
** NAME:        KvmRead2                                    NETMON
** SYNOPSIS:    int  KvmRead2
**              (
**                KVMOFST addr,     offset in the kernel space
**                void *buf,        user space recipient
**                unsigned nbytes   block length in bytes
**              )
** RELATED TO:  Kvm..()
** DESCRIPTION: transfer the contents of a kernel memory block
**              to user space. 'addr' must be a value retrieved
**              by means of KvmNlist().
** REMARKS:     - uses the global kernel access handle 'kvmH'
** RETURN TYPE: int
** EXIT STATUS:  0: OK
**              -1: read failure
*****************************************************************/

int KvmRead2( KVMOFST addr, void *buf, unsigned nbytes )
{
    register int done;

    done = kvm_read( kvmH, addr, (char*) buf, nbytes );

    if ( done == (int) nbytes )
    {
        kvmreadbytes += (double) nbytes;
        return 0;
    }
    else
    {
        kvmreaderrors++;
        return -1;
    }
}

int KvmRead3( KVMOFST addr, void *buf, unsigned nbytes )
{
    int done;

    off_t r;

    r = lseek( kvmF, (long) addr, SEEK_SET );
    if ( r == -1 )
         return -1;

    done = read( kvmF, (caddr_t) buf, (int) nbytes );
    if ( done == (int) nbytes )
    {
        kvmreadbytes += (double) nbytes;
        return 0;
    }
    else
    {
        kvmreaderrors++;
        return -1;
    }
}

/*****************************************************************
** NAME:        TimeOpen                                    NETMON
** SYNOPSIS:    void TimeOpen
**              (
**                void
**              )
** RELATED TO:  Time..()
** DESCRIPTION: initialise the timing system
** REMARKS:     - uses the kernel symbol '_time'
** RETURN TYPE: void
** EXIT STATUS: -
*****************************************************************/

void TimeOpen( void )
{
    /*
     * locate _time
     */

    KvmNlist( symlist, NSYM, __FILE__ );
    ofst_time = symlist[0].n_value;

    /*
     * set the time
     */

    timedone = timenew = timeold = 0;
    timebegin = TimeRead();
}

/*****************************************************************
** NAME:        TimeGet                                     NETMON
** SYNOPSIS:    void TimeGet
**              (
**                void
**              )
** RELATED TO:  Time..()
** DESCRIPTION: interval time fetch
** REMARKS:     - uses the kernel symbol '_time'
** RETURN TYPE: void
** EXIT STATUS: -
*****************************************************************/

void TimeGet( void )
{
     timeold  = timenew;
     timenew  = TimeRead() - timebegin;
     timedone = timenew - timeold;
}

/*****************************************************************
** NAME:        TimeRead                                    NETMON
** SYNOPSIS:    double TimeRead
**              (
**                void
**              )
** RELATED TO:  Time..()
** DESCRIPTION: read time elapsed since last boot
** REMARKS:     - uses the kernel symbol '_time'
** RETURN TYPE: double
** EXIT STATUS: time in seconds, with microsec resolution
*****************************************************************/

double TimeRead( void )
{
    static struct timeval uptime;

    KvmRead( ofst_time, &uptime, sizeof( uptime ));
    return  (((double)uptime.tv_sec) + 1e-6*((double)uptime.tv_usec));
}

/* end of netkern.c */

